import os
from dektools.file import sure_dir
from dektools.func import FuncAnyArgs
from dektools.design.node.base import NodeBase, NodeTypes, NodeManager
from ....svg import new_svg_manager


class NodeImage(NodeBase):
    pass


node_types = NodeTypes()


class ImageManager(NodeManager):
    types = node_types

    new_svg_manager = FuncAnyArgs(new_svg_manager)

    dump_ext = '.png'

    async def dump(self, path):
        def work(file, content):
            p = os.path.join(path, file + self.dump_ext)
            content.save(p)
            self.set_manager.load_path_item(p)

        sure_dir(path)
        if self.res_manager.has_fetch():
            async for node_name, name, params in self.res_manager.fetch_items():
                work(name, self.make(node_name, params=params))
        else:
            for name, item in self.entries():
                work(name, item)

    def entries(self, args=None, params=None, attrs=None):
        for name in self.res_manager.entry_names:
            yield name, self.make(name, args, params, attrs)

    def _translate_assign(self, s, params):
        if s is not None:
            return self.make(s, params=params)



from dektools.design.base import split_function
from dektools.dict import assign_list
from .base import NodeCanvas, NodeSvg, node_types
from .elements import *


@node_types.register
class FunctionNode(NodeCanvas):
    def make(self, args, params, attrs):
        _args, _params, body = self.res.content
        params = self.merge_params(_params, params)
        args = self.merge_args(_args, args, params)
        wh = assign_list([self.manager.default_width, self.manager.default_height], args)
        result = ""
        for key, value in body.items():
            _args, _params, _body = split_function(value)
            params = self.merge_params(_params, params)
            args = self.merge_args(_args, args, params)
            attrs = self.manager.translate_map(_body, params)
            name, trans = self.manager.parse_call_label(params, key, self)
            result += self.manager.make(
                name, args, params, attrs,
                init=lambda n: n.set_wh(*wh) if isinstance(n, NodeCanvas) else None,
                final=lambda n, r: f"""<g{trans}>{r}</g>""" if trans or n.need_g_wrapper else r
            )
        return result


@node_types.register
class SvgEmbedNode(NodeSvg):
    need_g_wrapper = True

    def make(self, args, params, attrs):
        if attrs:
            return self.manager.render_by_struct({'g': {**attrs, '+': self.res.content}}, params)
        return self.res.content


@node_types.register
class UnknownNode(NodeSvg):
    def make(self, args, params, attrs):
        return self.manager.render_by_struct({self.res.name: attrs}, params)



import typer
from dektools.shell import associate_console_script, associate_console_script_remove

app = typer.Typer(add_completion=False)


@app.command()
def assoc(delete: bool = typer.Option(False, "--delete", "-d")):
    from ..image.design.image.const import image_ignore_file
    from ..image.design.svg.const import svg_ignore_file
    args_list = [
        (image_ignore_file, __name__, 'ImageDesign', 'image design'),
        (svg_ignore_file, __name__, 'SvgDesign', 'image svg design'),
    ]
    for args in args_list:
        if delete:
            associate_console_script_remove(*args[:-1])
        else:
            associate_console_script(*args)



from PIL import Image
from dektools.design.res.base import ResBase, ResText, ResBytes
from dektools.ext.font import font_extensions
from dektools.ext.image import image_extensions


class SvgEmbedRes(ResText):
    ext = '.svg'


class SvgRes(ResBytes):
    ext = '.svg'


class ImageRes(ResBase):
    ext = image_extensions

    def load(self):
        return Image.open(self.path)


class FontRes(ResBase):
    ext = font_extensions


def get_fonts(res_manager):
    fonts = res_manager.get_res_map(FontRes.get_typed()) or {}
    if fonts:
        return {k: {'font_path': v.path} for k, v in fonts.items()}



from dektools.design.set.base import SetManager
from ..svg import load_svg
from .base.res import get_fonts
from .image import new_image_manager
from .svg import new_svg_manager


class SetMangerDump(SetManager):
    async def dump(self, path):
        await self.manager.dump(path)


class ImageSvgSetManager(SetMangerDump):
    managers = [new_image_manager, new_svg_manager]

    def trans_0_1(self, manager, node, result, args, params, attrs):
        return load_svg(node.manager.final(node, result), attrs.get('width'), attrs.get('height'),
                        fonts=get_fonts(manager.res_manager))


class SvgSetManager(SetMangerDump):
    managers = [new_svg_manager]



from dektools.design.res.base import ResBase, ResYaml, ResTypes
from deknetreq.design.apply import ResYamlFuncNet, RpRes, ResManagerNet
from ...base.res import SvgEmbedRes, FontRes

res_types = ResTypes()


@res_types.register
class SvgYamlRes(ResYamlFuncNet):
    ext = ['.svg', '.yaml']


res_types.register(SvgEmbedRes)
res_types.register(RpRes)
res_types.register(FontRes)


class SvgResManager(ResManagerNet):
    types = res_types

    res_cls_setup = SvgYamlRes



from dektools.design.res.base import ResBase, ResYaml, ResTypes
from deknetreq.design.apply import ResYamlFuncNet, RpRes, ResManagerNet
from ....psd import PsdCanvas
from ...base.res import SvgRes, ImageRes, FontRes

res_types = ResTypes()


@res_types.register
class ImageYamlRes(ResYamlFuncNet):
    ext = ['.image', '.yaml']


@res_types.register
class PsdRes(ResBase):
    ext = '.psd'

    def load(self):
        return PsdCanvas.load(self.path)


@res_types.register
class QrRes(ResYaml):
    ext = ['.qr', '.yaml']


res_types.register(SvgRes)
res_types.register(RpRes)
res_types.register(ImageRes)
res_types.register(FontRes)


class ImageResManager(ResManagerNet):
    types = res_types

    res_cls_setup = ImageYamlRes



from .const import svg_ignore_file
from .node.base import SvgManager
from .node import *
from .res import SvgResManager


def new_svg_manager(set_manager):
    return SvgManager(set_manager, SvgResManager(set_manager, [svg_ignore_file]))



from PIL import Image, ImageFilter


def gray(image: Image.Image):
    if image.has_transparency_data:
        return image.convert("LA")
    return image.convert("L")


def blur(image: Image.Image, radius=2):
    return image.filter(ImageFilter.GaussianBlur(radius))



from .const import image_ignore_file
from .node.base import ImageManager
from .node import *
from .res import ImageResManager


def new_image_manager(set_manager):
    return ImageManager(set_manager, ImageResManager(set_manager, [image_ignore_file]))



from PIL import Image
from skimage.io._plugins.pil_plugin import pil_to_ndarray, ndarray_to_pil
from dektools.module import get_module_attr
from dektools.design.base import split_function
from dektools.dict import assign
from ....svg import load_svg
from ....qr import make_qr
from ...base.res import get_fonts
from .base import NodeImage, node_types


@node_types.register
class FunctionNode(NodeImage):
    module_image_operations = (
        (f'{__name__}.operations.image', False),
        f'{__name__}.operations.array',
        'skimage',
        'numpy',
        ('PIL.ImageOps', False)
    )

    def make(self, args, params, attrs):
        _args, _params, body = self.res.content
        params = self.merge_params(_params, params)
        args = self.merge_args(_args, args, params)
        image = None
        for index, (name, value) in enumerate(body.items()):
            _args, _params, _body = split_function(value)
            params = self.merge_params(_params, params)
            args = self.merge_args(_args, args, params)
            attrs = self.manager.translate_map(_body, params)
            if index == 0:
                image = self.manager.make(name, args, params, attrs)
            else:
                name = name.rstrip('+')
                func = None
                is_na = True
                for m in self.module_image_operations:
                    if isinstance(m, str):
                        is_na = True
                    else:
                        m, is_na = m
                    try:
                        func = get_module_attr(f'{m}.{name}')
                        break
                    except (ModuleNotFoundError, AttributeError):
                        pass
                if func is None:
                    is_na = False
                    func = getattr(Image.Image, name, None)
                if func is None:
                    raise AttributeError(f"Can't find func: {name}")
                if is_na:
                    if isinstance(image, Image.Image):
                        image = pil_to_ndarray(image)
                else:
                    if not isinstance(image, Image.Image):
                        image = ndarray_to_pil(image)
                image = func(image, *args, **attrs)

        if not isinstance(image, Image.Image):
            return ndarray_to_pil(image)
        return image


@node_types.register
class SvgNode(NodeImage):
    def make(self, args, params, attrs):
        return load_svg(self.res.content, attrs.get('width'), attrs.get('height'),
                        fonts=get_fonts(self.manager.res_manager))


@node_types.register
class ImageNode(NodeImage):
    def make(self, args, params, attrs):
        return self.res.content


@node_types.register
class PsdNode(NodeImage):
    def make(self, args, params, attrs):
        self.res.content.update(attrs)
        return self.res.content.render()


@node_types.register
class QrNode(NodeImage):
    def make(self, args, params, attrs):
        return make_qr(**assign(attrs, self.manager.translate_map(self.res.content or {}, params)))



import os
from dektools.file import sure_dir, write_file
from dektools.design.node.base import NodeBase, NodeTypes, NodeManager
from dektools.func import FuncAnyArgs
from .....svg.utils.common import optimize_svg


class NodeSvg(NodeBase):
    need_g_wrapper = False


class NodeCanvas(NodeSvg):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.width = self.manager.default_width
        self.height = self.manager.default_height

    def set_wh(self, w, h):
        self.width = w
        self.height = h

    def pv(self, value, ratio=1):
        if isinstance(value, float):
            return self.width * value * ratio
        return value


class NodeElement(NodeCanvas):
    name = None
    spec = {}

    class Proxy:
        def __init__(self):
            self.params = {}

        def __getattr__(self, item):
            return self.params[item]

        def __setitem__(self, key, value):
            self.params[key] = value

    def new_proxy(self, attrs):
        proxy = self.Proxy()
        attrs = {**self.spec, **attrs}
        for k, v in attrs.items():
            if callable(v):
                v = FuncAnyArgs(v)(self.width, self.height, proxy)
            proxy[k] = v
        return proxy

    def make(self, args, params, attrs):
        return self.manager.render_by_struct(self.draw(self.new_proxy(attrs)), params)

    def draw(self, proxy):
        raise NotImplementedError


node_types = NodeTypes()


class SvgManager(NodeManager):
    types = node_types

    default_width = 1024
    default_height = default_width

    dump_ext = '.svg'

    async def dump(self, path):
        def work(file, content):
            p = os.path.join(path, file + self.dump_ext)
            write_file(p, b=content)
            self.set_manager.load_path_item(p)

        sure_dir(path)
        if self.res_manager.has_fetch():
            async for node_name, name, params in self.res_manager.fetch_items():
                work(name, self.make_svg(node_name, params=params))
        else:
            for name, item in self.entries():
                work(name, item)

    def entries(self, args=None, params=None, attrs=None):
        for name in self.res_manager.entry_names:
            yield name, self.make_svg(name, args, params, attrs)

    @staticmethod
    def final(node, result):
        return optimize_svg(
            f'<svg viewBox="0 0 {node.width} {node.height}" '
            f'xmlns="http://www.w3.org/2000/svg" '
            f'xmlns:xlink="http://www.w3.org/1999/xlink">{result}</svg>'
        ).encode('utf-8')

    def make_svg(self, name, args=None, params=None, attrs=None):
        return self.make(name, args, params, attrs, final=self.final)

    def render_by_struct(self, data, params):
        if isinstance(data, str):
            return data
        result = ""
        for label, attrs in data.items():
            tag, trans = self.parse_call_label(params, label)
            attrs = attrs.copy()
            children = attrs.pop('+', None)
            sa = "".join(f' {k}="{v}"' for k, v in attrs.items() if v not in ('', None))
            if children is None:
                result += f"<{tag}{trans}{sa}/>"
            else:
                result += f"<{tag}{trans}{sa}>{self.render_by_struct(children, params)}</{tag}>"
        return result

    def parse_call_label(self, params, label, node=None):
        def pv(tf, value):
            if node and tf in {'translate'}:
                return node.pv(value)
            return value

        def transform(s):
            for k, v in transform_map.items():
                if s.startswith(k):
                    tf = [str(pv(v, float(x) if '.' in x else int(x))) for x in s[len(k):].split(',')]
                    return f"{v}({','.join(tf)})"
            return ""

        transform_map = {'t': 'translate', 's': 'scale', 'sk': 'skew', 'r': 'rotate'}

        label = self.trans_var(label, params)

        kl = label.split(" ", 1)
        name = kl[0]
        if len(kl) == 2:
            items = [transform(x) for x in kl[1].split() if x and not x.startswith('-')]
        else:
            items = []
        if items:
            trans = f' transform="{" ".join(items)}"'
        else:
            trans = ""
        return name, trans
